from roistat import ab_tests
from http import cookies
import re
import random
import logging

ROISTAT_AB_COOKIE = "roistat_ab"

class ABTest:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        request.split_tests = Tests(request)

        response = self.get_response(request)

        request.split_tests._write_cookies(response)
        return response


class Tests:
    def __init__(self, request):
        logging.info('Initialising split tests module')

        self._request = request
        self._cookies = {}
        self._tests_values = {}
        self._is_resubmit_required = False

        if not self._init_tests():
            return

        if self._user_is_seo_bot():
            logging.info('User is bot. Skip')
            return

        self._init_from_general_cookie()
        self._init_from_tests()

    def _init_tests(self):
        tests = ab_tests.tests

        if type(tests) is not tuple:
            logging.error('Tests are not tuple type. Abort')
            return False
        elif len(tests) == 0:
            logging.info('No tests configured. Skip initialisation')
            return False

        self._tests = {}
        for test in tests:
            if type(test) is not dict:
                logging.error('Test is not a dictionary type. Abort test initialisation')
                continue

            if 'id' not in test:
                logging.error('Test does not have id key. Abort test initialisation')
                continue

            if 'variants' not in test:
                logging.error('Test does not have variants key. Abort test initialisation')
                continue

            self._tests[test['id']] = test

        return True

    def values(self):
        return TestsValues(self)

    def _user_is_seo_bot(self):
        if "HTTP_USER_AGENT" not in self._request.META:
            return False

        match = re.search("#|bot|crawl|slurp|spider#", self._request.META["HTTP_USER_AGENT"].lower())
        return match is not None

    def _init_from_general_cookie(self):
        logging.info('Initialising tests values from general cookie')
        general_cookie = self._get_cookie(ROISTAT_AB_COOKIE)
        if general_cookie is None:
            logging.info('No general cookie found')
            return

        cookies = general_cookie.split(',')
        for cookie in cookies:
            data = cookie.split(':')
            if len(data) != 2:
                logging.error('Strange test value '+str(data))
                continue

            self._tests_values[data[0]] = data[1]

    def _init_from_tests(self):
        logging.info('Initialising tests')
        for test_id, test in self._tests.items():
            if test_id in self._tests_values:
                logging.info('Test value has been set from general cookie, continue')
                continue

            test_cookie = self._get_cookie(test_id)
            if test_cookie is not None and self._is_valid_test_cookie(test_cookie, test):
                self._set_test_value(test, test_cookie)
                logging.info('Test value set from test cookie')
                continue

            if 'active' in test and test['active'] == False:
                logging.info('Inactive test, skip initial setup')
                continue

            self._get_value_from_test(test)

    def _get_value_from_test(self, test):
        logging.info('Get test value')
        if "method" in test:
            logging.info('Value received from method')
            value = test['method'](test)
        else:
            logging.info('Value received randomly')
            value = self._calc_random_variant(test)

        self._set_test_value(test, value)
        return value

    def _get_cookie_flag(self, cookie):
        if cookie not in self._request.COOKIES:
            return False
        if self._request.COOKIES[cookie] != '':
            return int(self._request.COOKIES[cookie]) > 0
        return False

    @staticmethod
    def _is_valid_test_cookie(value, test):
        for variant in test['variants']:
            if value == variant['id']:
                return True
        return False

    @staticmethod
    def _calc_random_variant(test):
        return random.choice([variant['id'] for variant in test['variants']])

    def _set_test_value(self, test, value):
        self._tests_values[test['id']] = value
        self._set_cookie(test['id'], value)
        self._is_resubmit_required = True

    def _set_cookie(self, cookie, value):
        self._cookies[cookie] = value

    def _write_cookies(self, response):
        if self._is_resubmit_required:
            self._set_roistat_cookie()

        for cookie, value in self._cookies.items():
            response.set_cookie(cookie, str(value), max_age=3110400)

    def _get_cookie(self, cookie):
        if cookie not in self._request.COOKIES:
            return None
        return self._request.COOKIES[cookie]

    def _set_roistat_cookie(self):
        if len(self._tests_values) == 0:
            return

        cookies = []
        for test_id, value in self._tests_values.items():
            cookies.append(str(test_id)+':'+str(value))

        self._set_cookie(ROISTAT_AB_COOKIE, ','.join(cookies))


class TestsValues:
    def __init__(self, tests):
        self._tests = tests

    def __getattr__(self, name):
        if name in self._tests._tests_values:
            return self._tests._tests_values[name]

        if name in self._tests._tests:
            return self._tests._get_value_from_test(self._tests._tests[name])

        return None
